import { describe, it, expect, beforeEach } from 'vitest'
import { v4 } from 'uuid'
import { createPinia, setActivePinia } from 'pinia'

import { useMetaSave } from '@/composables/useMetaSave'

import { useFormStore, useSubmissionStore } from '@vue-unistra/formbuilder'
import { LAYOUT_COMPONENTS } from '@/stores/template'

import type { Metadata } from '#/item'
import type { LevelField, ReferentialField } from '#/referential'

describe('useMetaSave', () => {
  setActivePinia(createPinia())

  const submitStore = useSubmissionStore()
  const formStore = useFormStore()

  function insertField(component: string, parentId= 'root') {
    const id = v4()
    formStore.fields[id] = {
      component,
      exist: true,
      id,
      parentId,
      visible: true,
    }
    return id
  }

  beforeEach(() => {
    submitStore.$reset()
    formStore.$reset()
  })
  describe('inputFields', () => {
    it('should exclude layout fields defined in template store', () => {
      LAYOUT_COMPONENTS.forEach(layoutComponent => {
        insertField(layoutComponent)
      })
      insertField('randomInputField')
      const {inputFields} = useMetaSave()
      expect(inputFields.value).toHaveLength(1)
      expect(inputFields.value[0].component).toBe('randomInputField')
    })
  })
  describe('storedMetadata', () => {
    let fieldId: string

    function getMetadata(): Metadata {
      const { storedMetadata } = useMetaSave()
      return storedMetadata.value[0]
    }

    beforeEach(() => {
      fieldId = insertField('randomInputField')
    })

    it('should use its id as default name', () => {
      const fieldMetadata = getMetadata()
      expect(fieldMetadata.name).toEqual(fieldId)
    })
    it('should return null value with no submitted value', () => {
      const fieldMetadata = getMetadata()
      expect(fieldMetadata.value).toBeNull()
    })
    it('should get value from submission store if set', () => {
      const inputValue = "myValue"
      submitStore.submission[fieldId] = inputValue
      const fieldMetadata = getMetadata();
      expect(fieldMetadata.value).toBe(inputValue)
    })
    it('should not add other keys if not specific field', () => {
      const fieldMetadata = getMetadata();
      [
        'isOptions',
        'levelId',
        'multiple',
        'parentId',
        'referentialId',
        'repeatable',
      ].forEach(specificKey => {
        expect(Object.hasOwn(fieldMetadata, specificKey)).toBe(false)
      })
    })
    it('should set multiple key to true for such field', () => {
      formStore.fields[fieldId].multiple = true
      const fieldMetadata = getMetadata();
      expect(fieldMetadata.multiple).toBe(true)
    })
    it('should set repeatable key to true for such field', () => {
      formStore.fields[fieldId].repeatable = true
      const fieldMetadata = getMetadata();
      expect(fieldMetadata.repeatable).toBe(true)
    })
    it('should set isOption key to true if value come from option list', () => {
      formStore.fields[fieldId].options = []
      const fieldMetadata = getMetadata();
      expect(fieldMetadata.isOption).toBe(true)
    })
    it('should set levelId & referentialId for referential value', () => {
      const referentialId = v4()
      const levelId = v4()

      const levelSelectId = insertField('LevelSelect', fieldId)
      const referentialField: ReferentialField = formStore.fields[fieldId] as ReferentialField
      referentialField.component = 'ReferentialComponent'
      referentialField.referentialId = referentialId;
      (formStore.fields[levelSelectId] as LevelField).levelId = levelId

      const fieldMetadata = getMetadata();
      expect(fieldMetadata.levelId).toBe(levelId)
      expect(fieldMetadata.referentialId).toBe(referentialId)
    })
  })
})
