import { describe, it, beforeEach, expect, vi } from 'vitest'
import {Token, Authority} from '@/services/authService'
import {tokenStorage as storage } from '@/services/storageService'
import jwt from 'jwt-simple'
import type { Mock } from 'vitest'

const generateToken = (secondDeltaFromNow: number): string => jwt.encode(
    { exp: Math.trunc(Date.now() / 1000) + secondDeltaFromNow },
    'key'
  )

describe('Token class', () => {
  let validToken: string, expiredToken: string, willSoonExpireToken: string
  beforeEach(() => {
    validToken = generateToken(10)
    expiredToken = generateToken(-1)
    willSoonExpireToken = generateToken(2)
  })
  describe('constructor', () => {
    it('sould init expired token from a not base 64 encoded string', () => {
      const token = new Token('')
      expect(token.timeToExpiration).toEqual(-1)
    })
    it('sould init expired token from base 64 encoded not valid json string', () => {
      const jwtBody = btoa('][')
      const token = new Token(`h.${jwtBody}.f`)
      expect(token.timeToExpiration).toEqual(-1)
    })
    it('sould init specific body from a not base 64 encoded string', () => {
      const token = new Token('')
      expect(token.toString()).toEqual('invalid base64 token')
    })
    it('sould init expired token from jwt without expiration property', () => {
      const token = new Token(jwt.encode({iss: 'me'}, 'key'))
      expect(token.timeToExpiration).toEqual(-1)
    })
    it('sould init expired token from expired jwt', () => {
      const token = new Token(expiredToken)
      expect(token.timeToExpiration).to.be.lt(0)
    })
    it('sould init non expired token from non expired jwt', () => {
      const token = new Token(validToken)
      expect(token.timeToExpiration).to.be.gt(0)
    })
  })
  describe('getters', () => {
    it('should access body through toString method', () => {
      const token = new Token(validToken)
      expect(`${token}`).toEqual(validToken)
    })
    describe('timeToExpiration', () => {
      it('should be an integer', () => {
        const token = new Token(validToken)
        expect(Number.isSafeInteger(token.timeToExpiration)).toBe(true)
      })
      it('should be expressed in millisecond', () => {
        const token = new Token(validToken)
        expect(token.timeToExpiration).to.be.gte(8 * 1000)
        expect(token.timeToExpiration).to.be.lt(10 * 1000)
      })
    })
    describe('isUp', () => {
      it('should be true if token wont expire within 5 seconds', () => {
        const token = new Token(validToken)
        expect(token.isUp).toBe(true)
      })
      it('should be false if token will expire within 5 seconds', () => {
        const token = new Token(willSoonExpireToken)
        expect(token.isUp).toBe(false)
      })
      it('should be false if token has already expired', () => {
        const token = new Token(expiredToken)
        expect(token.isUp).toBe(false)
      })
    })
  })
})

describe('Authority class', () => {
  let authority: Authority
  let validToken: string, expiredToken: string, willSoonExpireToken: string
  beforeEach(() => {
    authority = new Authority()
    validToken = generateToken(10)
    expiredToken = generateToken(-1)
    willSoonExpireToken = generateToken(2)
    storage.clear()
  })
  describe('getters', () => {
    let expired: Token, soon: Token, valid: Token
    beforeEach(() => {
      expired = new Token(expiredToken)
      soon = new Token(willSoonExpireToken)
      valid = new Token(validToken)
    })
    describe('isActive', () => {
      it('should be false if missing token', () => {
        authority['access'] = valid
        expect(authority.isActive).toBe(false)
      })
      it('should be false if access will expire or is expired', () => {
        authority.updatePair(soon, valid)
        expect(authority.isActive).toBe(false)
      })
      it('should be false if refresh will expire or is expired', () => {
        authority.updatePair(valid, expired)
        expect(authority.isActive).toBe(false)
      })
      it('should be true with valid access and refresh', () => {
        authority.updatePair(valid, valid)
        expect(authority.isActive).toBe(true)
      })
    })
    describe('shouldRefresh', () => {
      it('should be false if missing refresh', () => {
        authority['access'] = valid
        expect(authority.shouldRefresh).toBe(false)
      })
      it('should be false if refresh will expire or is expired', () => {
        authority.updatePair(expired, soon)
        expect(authority.shouldRefresh).toBe(false)
      })
      it('should be false if access is already up', () => {
        authority.updatePair(valid, valid)
        expect(authority.shouldRefresh).toBe(false)
      })
      it('should be true with valid refresh and expired access', () => {
        authority.updatePair(expired, valid)
        expect(authority.shouldRefresh).toBe(true)
      })
      it('should be true with valid refresh and missing access', () => {
        authority['refresh'] = valid
        expect(authority.shouldRefresh).toBe(true)
      })
    })
  })
  describe('methods', () => {
    let expired: Token, valid: Token
    beforeEach(() => {
      expired = new Token(expiredToken)
      valid = new Token(validToken)
    })
    describe('update methods', () => {
      it('setAuthHeader should set Authorization with access', () => {
        authority['access'] = new Token(validToken)
        authority.setAuthHeader()
        expect(Object.prototype.hasOwnProperty.call(
            authority['instance'].defaults.headers.common, 'Authorization'
        )).toBe(true)
        expect(authority['instance'].defaults.headers.common.Authorization)
            .toEqual(`Bearer ${validToken}`)
      })
      it('setAutoRefresh should call token refresh before access expiration', async () => {
        authority.tokenRefresh = vi.fn()
        authority['access'] = new Token(generateToken(3))
        authority.setAutoRefresh()
        const testTimeout = () => new Promise(resolve => {
          setTimeout(resolve, 1000)
        }).then(() => {
          expect(authority.tokenRefresh).toHaveBeenCalledOnce()
        })
        await testTimeout()
      })
      describe('initAutoAuth', () => {
        beforeEach(() => {
          authority.setAuthHeader = vi.fn()
          authority.setAutoRefresh = vi.fn()
          authority['access'] = new Token(validToken)
          expect(authority.available).toBe(false)
          authority.initAutoAuth()
        })
        it('should call setAuthHeader', () => {
          expect(authority.setAuthHeader).toHaveBeenCalledOnce()
        })
        it('should call setAutoRefresh', () => {
          expect(authority.setAutoRefresh).toHaveBeenCalledOnce()
        })
        it('should change service availability status', () => {
          expect(authority.available).toBe(true)
        })
      })
      describe('token update', () => {
        beforeEach(() => {
          authority.initAutoAuth = vi.fn()
          authority['access'] = expired
          authority['refresh'] = expired
        })
        describe('updateAccess', () => {
          beforeEach(async () => {
            await authority.updateAccess(validToken)
          })
          it('should update access token with given string', () => {
            expect((authority['access'] as Token).toString()).toEqual(validToken)
          })
          it('should update access in storage', () => {
            expect(storage.get('access')).toEqual(validToken)
          })
          it('should call initAutoAuth', () => {
            expect(authority.initAutoAuth).toHaveBeenCalledOnce()
          })
        })
        describe('updatePair', () => {
          let refreshToken: Token, accessToken: Token
          beforeEach(() => {
            accessToken = new Token(validToken)
            refreshToken = new Token(generateToken(60))
            authority.updatePair(accessToken, refreshToken)
          })
          it('should update access with given token', () => {
            expect(authority['access']).to.eql(accessToken)
          })
          it('should update refresh with given token', () => {
            expect(authority['refresh']).to.eql(refreshToken)
          })
          it('should call initAutoAuth', () => {
            expect(authority.initAutoAuth).toHaveBeenCalledOnce()
          })
        })
      })
      describe('tokenRefresh', () => {
        const refresh = generateToken(10000)
        beforeEach(() => {
          authority['refresh'] = new Token(refresh)
          authority['instance'].post = vi.fn()
          authority.updateAccess = vi.fn()
        })
        it('should post refresh to login api', async() => {
          (authority['instance'].post as Mock).mockResolvedValueOnce({data: { access: validToken } })
          await authority.tokenRefresh()
          expect(authority['instance'].post).toHaveBeenCalledOnce()
          expect(authority['instance'].post).toHaveBeenLastCalledWith(
              'login/refresh/',
              { refresh }
          )
        })
        it('should update access with retrieve data', async () => {
          (authority['instance'].post as Mock).mockResolvedValueOnce({data: { access: validToken } })
          await authority.tokenRefresh()
          expect(authority.updateAccess).toHaveBeenNthCalledWith(1,validToken)
        })
        describe('failure handling', () => {
          const apiError = new Error('reason')
          beforeEach(() => {
            (authority['instance'].post as Mock).mockRejectedValue(apiError)
          })
          it('should throw an error', async () => {
            const errorTest = () => authority.tokenRefresh().catch(
                (error: Error) => expect(error).toEqual(apiError)
            )
            await errorTest()
          })
          it('should let service unavailable', async () => {
            const errorTest = () => authority.tokenRefresh().catch(() => {})
            await errorTest()
            expect(authority.available).to.be.equal(false, 'service unavailable')
          })
          it('should not try to update access', async () => {
            const errorTest = () => authority.tokenRefresh().catch(() => {})
            await errorTest()
            expect(authority.updateAccess).toHaveBeenCalledTimes(0)

          })
          it('should try three time to refresh before fail', async () => {
            const refreshSpy = vi.spyOn(authority, 'tokenRefresh')
            const errorTest = () => authority.tokenRefresh().catch(() => {})
            await errorTest()
            expect(refreshSpy).toHaveBeenCalledTimes(3)
          })//.timeout(5000)
        })
      })
    })
    describe('logOut', () => {
      it('should delete Authorization header', () => {
        authority['access'] = valid
        authority.setAuthHeader()
        authority.logOut()
        expect(Object.prototype.hasOwnProperty.call(
            authority['instance'].defaults.headers.common,
            'Authorization'
        )).toBe(false)
      })
      it('should remove access and refresh keys from user\'s storage', () => {
        storage.dump({access: validToken, refresh: validToken})
        authority.logOut()
        expect(storage.get('access', 'missing'), 'access removed').toEqual('missing')
        expect(storage.get('refresh', 'no here'), 'refresh removed').toEqual('no here')
      })
      it('should nullify own access and refresh', () => {
        authority['refresh'] = new Token(validToken)
        authority.logOut()
        expect(authority['access']).to.be.null
        expect(authority['refresh']).to.be.null
      })
      it('should set availability to false', () => {
        authority.available = true
        authority.logOut()
        expect(authority.available).toBe(false)
      })
      it('should clear access refresh timeout', async () => {
        authority.tokenRefresh = vi.fn()
        authority['access'] = new Token(generateToken(6))
        authority.setAutoRefresh()
        authority.logOut()
        const testTimeout = () => new Promise(resolve => {
          setTimeout(resolve, 1500)
        }).then(() => {
          expect(authority.tokenRefresh).toHaveBeenCalledTimes(0)
        })
        await testTimeout()
      })
    })
    describe('initToken', () => {
      beforeEach(() => {
        authority.updatePair = vi.fn()
        authority.tokenRefresh = vi.fn().mockResolvedValueOnce({})
        authority.available = true
      })
      it('should set service to unavailable if no token stored', async () => {
        await authority.initToken()
        expect(authority.available).toBe(false)
      })
      it('should set service to unavailable if invalid refresh stored', async () => {
        storage.set('access', validToken)
        storage.set('refresh', expiredToken)
        await authority.initToken()
        expect(authority.available).toBe(false)
      })
      it('should refresh access with only valid refresh', async () => {
        storage.set('access', expiredToken)
        storage.set('refresh', validToken)
        await authority.initToken()
        expect(authority.tokenRefresh).toHaveBeenCalledOnce()
        expect(authority['refresh']).toEqual(new Token(validToken))
      })
      it('should update pair if both valid', async () => {
          const access = generateToken(10)
          const refresh = generateToken(20)
          storage.set('access', access)
          storage.set('refresh', refresh)
          await authority.initToken()
          expect(authority.updatePair).toHaveBeenCalledOnce()
          expect(authority.updatePair).toHaveBeenLastCalledWith(
              new Token(access), new Token(refresh)
          )
        })
    })
    describe('logIn', () => {
      let tokenResponse: {access: string, refresh: string}
      beforeEach(() => {
        tokenResponse = {
          access: validToken,
          refresh: generateToken(84),
        }
        authority['instance'].post = vi.fn().mockResolvedValueOnce({data: tokenResponse })
        authority.updatePair = vi.fn()

      })
      describe('with no valid credentials', () => {
        let errorResponse: Error
        beforeEach(async () => {
          const errorTest = () => authority.logIn({}).catch(
              error => errorResponse = error
          )
          await errorTest()
        })
        it('should not call api', () => {
          expect(authority['instance'].post).toHaveBeenCalledTimes(0)
        })
        it('should return a specific Type error', () => {
          expect(errorResponse instanceof TypeError, 'class check').toBe(true)
          expect(errorResponse.message).toEqual('credentials provided are not valid')
        })
      })
      describe('valid credentials', () => {
        const ticketCredentials = { ticket: 'toRide' }
        it('should be posted to token login api if a ticket is provided', async () => {
          await authority.logIn(ticketCredentials)
          expect(authority['instance'].post).toHaveBeenCalledOnce()
          expect(authority['instance'].post).toHaveBeenLastCalledWith(
              'login/token/',
              ticketCredentials
          )
        })
        it('should call updatePair with received tokens', async () => {
          await authority.logIn(ticketCredentials)
          expect(authority.updatePair).toHaveBeenCalledOnce()
          expect(authority.updatePair).toHaveBeenLastCalledWith(
              new Token(tokenResponse['access']),
              new Token(tokenResponse['refresh'])
          )
        })
        it('should fill user\'s storage with received token', async () => {
          await authority.logIn(ticketCredentials)
          expect(storage.get('access'), 'access stored')
              .toEqual(tokenResponse['access'])
          expect(storage.get('refresh'), 'refresh stored')
              .toEqual(tokenResponse['refresh'])
        })
        it('should throw api error response if post failed', async () => {
          const apiError = new Error('reason');
          (authority['instance'].post as Mock).mockRejectedValueOnce(apiError)
          const errorTest = () => authority.logIn(ticketCredentials).catch(
              error => expect(error).to.eql(apiError)
          )
          await errorTest()
        })
        it('should be posted to local login api if username & password are provided', async () => {
          const creds = { username: 'random', password: 'secret' }
          await authority.logIn(creds)
          expect(authority['instance'].post).toHaveBeenCalledOnce()
          expect(authority['instance'].post).toHaveBeenLastCalledWith(
              'login/local/',
              creds
          )
        })
        it('should be posted to password reset login api if relevant infos are provided', async () => {
          const creds = {
            newPassword: 'secret',
            newPassword2: 'secret',
            token: 'tok',
            rid: '456e-7a56',
          }
          await authority.logIn(creds)
          expect(authority['instance'].post).toHaveBeenCalledOnce()
          expect(authority['instance'].post).toHaveBeenLastCalledWith(
              'login/password-reset/signin/',
              creds
          )
        })
      })
    })
  })
})
