import { describe, it, beforeEach, expect } from 'vitest'
import { NamedStorage } from '@/services/storageService'
import { storageMock } from '../../fixtures/localStorage'

describe('Storage class', () => {
  let storage: NamedStorage
  beforeEach(() => {
    storage = new NamedStorage('prefix', storageMock())
  })
  describe('get', () => {
    beforeEach(() => {
      storage.storage.setItem('key', 'value')
      storage.storage.setItem('prefix/key', 'prefixedValue')
    })
    it('should retrieve stored value from prefixed key', () => {
      expect(storage.get('key')).to.equal('prefixedValue')
    })
    it('should return empty string if key not stored', () => {
      expect(storage.get('random')).to.equal("")
    })
    it('should return fallback if given and key not stored', () => {
      expect(storage.get('random', 'des choux')).to.equal('des choux')
    })
  })
  describe('set', () => {
    beforeEach(() => {
      storage.set('key', 'value')
    })
    it('should add a prefixed key', () => {
      expect(storage.storage.key(0)).to.equal('prefix/key')
    })
    it('should add the given value', () => {
      expect(storage.storage.getItem(storage.storage.key(0) as string)).to.equal('value')
    })
    it('should override existing value', () => {
      storage.set('key', 'anotherValue')
      expect(storage.get('key')).to.equal('anotherValue')
    })
  })
  describe('del', () => {
    beforeEach(() => {
      storage.set('stay', 'a')
      storage.set('delete', 'b')
      storage.del('delete')
    })
    it('should remove value of existing key in storage', () => {
      expect(storage.get('delete', ''), 'value removed').toEqual('')
    })
    it('should remove an existing key of storage', () => {
      expect(Object.prototype.hasOwnProperty.call(storage.storage, 'delete'))
          .to.be.false
    })
    it('should remove let other key unchanged', () => {
      expect(storage.get('stay', ''), 'value unchanged')
          .to.equal('a')
    })
  })
  describe('update', () => {
    let state: Record<string, string>
    beforeEach(() => {
      state = {a: 'A', b: 'B'}
      storage.set('a', 'a')
      storage.update(state)
    })
    it('should update key if found in storage', () => {
      expect(state.a).to.equal('a')
    })
    it('should not update key if not in storage', () => {
      expect(state.b).to.equal('B')
    })
  })
})

