import {describe, it, beforeEach, afterEach, expect, vi} from 'vitest'
import { createPinia, setActivePinia } from 'pinia'
import {useDialogStore} from '@/stores/dialog'
import { authAxios } from '@/services/authService'
import { notifyError} from '@/services/alertService'
import { t } from '@/plugins/i18n'
import type { Mocked } from 'vitest'
import type { Authority } from '@/services/authService'
import type { AppNotification } from '../../../types/notification'

vi.mock(
    '@/services/alertService',
    () => ({
      notifyError: vi.fn()
    })
)

vi.mock(
    '@/services/authService',
    () => ({
      authAxios: {
        get: vi.fn(),
        patch: vi.fn(),
      }
    })
)

const mockedAxios: Mocked<Authority> = vi.mocked(authAxios, true)

describe('store/dialog', () => {
  setActivePinia(createPinia())
  const store = useDialogStore()
  afterEach(() => {
    vi.restoreAllMocks()
  })
  describe('actions', () => {
    const axiosError = new Error('api error')
    const testError = (apiMethod: 'get' | 'patch', storeAction: Function, translationKey: string, ...params: any[]) =>
        async () => {
          mockedAxios[apiMethod].mockRejectedValueOnce(axiosError)
          await storeAction(...params)
          expect(notifyError).toHaveBeenCalledOnce()
          expect(notifyError).toHaveBeenLastCalledWith(
              t(translationKey),
              axiosError
          )
        }
    beforeEach(() => {
      store.$reset()
    })
    describe('getNotifications', () => {
      it('should call notifications api for asked page', async () => {
        await store.getNotifications(2)
        expect(mockedAxios.get, 'api called').toHaveBeenCalledOnce()
        expect(mockedAxios.get).toHaveBeenLastCalledWith(
            'notifications/unread/',
            {params: {page: 2}}
        )
      })
      it('should ask default first page', async () => {
        await store.getNotifications()
        expect((mockedAxios.get.mock.lastCall as any[])[1].params.page).toEqual(1)
      })

      it(
          'should display error on failure',
          testError(
            'get',
            store.getNotifications,
            'notification.error.couldNotFetchNotifications'
          )
      )
      it('should update notify state with received data', async () => {
        const results = [{}]
        const receivedData = {
          previousPage: 1,
          nextPage: 3,
          count: 4,
        }
        mockedAxios.get.mockResolvedValueOnce({data: {...receivedData, results}})
        await store.getNotifications(2)
        expect(store.notify).toEqual({
          ...receivedData, notifications: results
        })
      })
    })
    describe('markNotificationRead', () => {
      it('should call mark read notifications api', () => {
        store.markNotificationRead(45)
        expect(mockedAxios.patch, 'api check')
            .toHaveBeenNthCalledWith(1, 'notifications/unread/45/mark_read/')
      })
      it(
        'should display error on failure',
        testError(
          'patch',
          store.markNotificationRead,
          'notification.error.markRead'
        )
      )
      it('should update notify state with received data', async () => {
        const results = [{}]
        const receivedData = {
          previousPage: 1,
          nextPage: 3,
          count: 4,
        }
        mockedAxios.patch.mockResolvedValueOnce({data: {...receivedData, results}})
        await store.markNotificationRead(2)
        expect(store.notify).toEqual({
          ...receivedData, notifications: results
        })
      })
    })
    describe('markAllRead', () => {
      it('should call mark all notifications read api', () => {
        store.markAllRead()
        expect(mockedAxios.patch)
            .toHaveBeenNthCalledWith(1, 'notifications/mark_all_read/')
      })
      it(
        'should display error on failure',
        testError(
          'patch',
          store.markAllRead,
          'notification.error.markAllRead'
        )
      )
      it('should reset notify state', async () => {
        store.notify = {
          previousPage: null,
          nextPage: null,
          count: 5,
          notifications: [{}, {}, {}, {}, {}] as AppNotification[]
        }
        mockedAxios.patch.mockResolvedValueOnce({status: 204})
        await store.markAllRead()
        expect(store.notify).toEqual({
          previousPage: null,
          nextPage: null,
          count: 0,
          notifications: []
        })
      })
    })
  })
})
