import { it, expect, describe, beforeEach, afterEach, vi } from 'vitest'
import { nextTick } from 'vue'
import { useRouter } from 'vue-router'
import { createPinia, setActivePinia } from 'pinia'
import { mount } from '@vue/test-utils'
import { Quasar } from 'quasar'
import { i18n, t } from '@/plugins/i18n'
import LoginPage from '@/views/signin/LoginPage.vue'
import { useContribStore } from '@/stores/contrib'

import type { ComputedRef } from 'vue'
import type { DOMWrapper } from '@vue/test-utils'
import type { TestWrapper } from '../types'

type Credentials = {
  username: string;
  password: string
}

type LocalConnectionData = {
  name: string;
  color: string;
  icon: string;
  show: boolean;
  valid: boolean;
  credentials: Credentials
}

type PropsType = {
  from?: string;
  toPath?: string
}

vi.mock('vue-router', () => ({
  useRouter: () => ({
    push: vi.fn(),
    replace: vi.fn(),
    resolve: vi.fn().mockImplementation(() => ({ href: 'href' }))
  })
}))

const router = useRouter()

// remove q-page wrapper from component before test
describe.skip('login', () => {
  setActivePinia(createPinia())
  const store = useContribStore()

  let wrapper: TestWrapper<{providers: any[], localConnect: ComputedRef<LocalConnectionData>}>
  function mountLogin(props: PropsType, shallow=false) {
    wrapper = mount(LoginPage,
        {
          props,
          global: {plugins: [i18n, Quasar]},
          shallow,
        })
  }
  afterEach(() => {
    wrapper.unmount()
  })
  describe('Component init', () => {
    it('should build providers url', () => {
      mountLogin({}, false)
      expect(wrapper.exists()).toBe(true)
      for (const provider of wrapper.vm.providers) {
        expect(provider.url).toContain(`login/${provider.name}/?redirect=`)
      }
    })
  })
  describe('structure', () => {
    beforeEach(() => {
      mountLogin({from: '/'})
    })
    it('should contains a cancel button in page bottom', () => {
      const cancelButton = wrapper.find('#content > button')
      expect(cancelButton.exists()).toBe(true)
      expect(cancelButton.text(), 'cancel label').toEqual(t('btn.cancel'))
    })
    it('should contains a local connection button in dedicated section', () => {
      const cancelButton = wrapper.find('#content > section > button')
      expect(cancelButton.exists()).toBe(true)
      expect(cancelButton.text()).toEqual('pount')
    })
    it('should contains a button link per provider', () => {
      expect(wrapper.findAll(`a.provider`)).toHaveLength(1)
    })
  })
  /** not tested */
  describe.skip('methods', () => {
    let validCredentials: Credentials
    beforeEach(() => {
      mountLogin({from: '/'})
      validCredentials = {
        username: 'keybord@azer.ty',
        password: 'itsavalidpassword'
      }
    })
    describe('cancel logIn', () => {
      it('should route to cancel route', async () => {
        const cancelButton = wrapper.find('button')
        await cancelButton.trigger('click')
        await nextTick()
        expect(router.push).toHaveBeenNthCalledWith(1, {path: '/'})
      })
    })
    describe('resetCredentials', () => {
      let localResult: LocalConnectionData
      beforeEach(() => {
        wrapper.vm.localConnect = {
          name: 'pount',
          color: '#aa5f0d',
          icon: 'mdi-school',
          credentials: validCredentials,
          show: true,
          valid: false
        }
        wrapper.vm.resetCredentials()
        localResult = wrapper.vm.localConnect.value as LocalConnectionData
      })
      it('should reset form valid state', function () {
        expect(localResult.valid).to.be.true
      })
      it('should reset form display state', function () {
        expect(localResult.show).to.be.false
      })
      it('should reset form fields data', function () {
        expect(localResult.credentials)
            .to.be.eql({
          username: '',
          password: ''
        })
      })
    })
    describe('local connect button', () => {
      const resetStub = vi.fn()
      let localButton: DOMWrapper<HTMLButtonElement>
      beforeEach(async () => {
        wrapper.vm.resetCredentials = resetStub
        localButton = wrapper.findAll('button').at(0) as DOMWrapper<HTMLButtonElement>
        await localButton.trigger('click')
        await nextTick()
      })
      it('should open and close logIn form', async () => {
        expect(wrapper.findAll('form')).to.have.lengthOf(1)
        await localButton.trigger('click')
        await wrapper.vm.$nextTick()
        expect(wrapper.findAll('form')).to.have.lengthOf(0)
      })
      it('should reset form fields on close', async () => {
        await localButton.trigger('click')
        expect(resetStub).toHaveBeenCalledOnce()
      })
      describe('signIn button', () => {
        let signInButton: DOMWrapper<HTMLButtonElement>
        const validateMock = vi.fn()
        beforeEach(() => {
          signInButton = wrapper.find('form button')
          store.logIn = vi.fn()
          wrapper.vm.$refs.form.validate = validateMock
        })
        it('should not dispatch logIn with no creds', async () => {
          validateMock.mockReturnValue(false)
          await signInButton.trigger('click')
          expect(store.logIn).toHaveBeenCalledTimes(0)
        })
        describe('with valid credentials', () => {
          beforeEach(async () => {
            router.replace = vi.fn()

            validateMock.mockReturnValue(true)
            wrapper.vm.localConnect.valid = true
            wrapper.vm.localConnect.credentials = validCredentials
            await wrapper.vm.$nextTick()
          })
          it('should call logIn method with credentials and redirect path', async () => {
            await signInButton.trigger('click')
            expect(store.logIn).toHaveBeenNthCalledWith(
                1,
                { ...validCredentials, route: {path: '/dashboard/projects'}}
            )
          })
          it('should reset credentials', async () => {
            await signInButton.trigger('click')
            await nextTick()
            expect(resetStub).toHaveBeenCalledOnce()
          })
        })
      })
      describe('form field validation', () => {
        it('should valid creds', async () => {
          wrapper.vm.localConnect.valid = false
          wrapper.vm.localConnect.credentials.username = 'keybord@azer.ty'
          wrapper.vm.localConnect.credentials.password = 'itsavalidpassword'
          await wrapper.vm.$nextTick()
          wrapper.vm.$refs.form.validate()
          await wrapper.vm.$nextTick()
          expect(wrapper.vm.localConnect.valid).to.equal(true, 'valid credentials')
        })
      })
    })
  })
})
